from .collapse_literals import collapse_literals
from . import core
from .deindex import deindex
from .inline import inline
from .unroll import unroll
# from .cleanup import cleanup
from .lift import lift


import ast
import inspect
import logging
import functools

import astor
from miniutils.magic_contract import safe_new_contract

log = logging.getLogger(__name__.split('.')[0])


def _is_iterable(x):
    return hasattr(x, '__iter__')

def _is_indexable(x):
    return hasattr(x, '__getitem__')


safe_new_contract('iterable', _is_iterable)
safe_new_contract('indexable', _is_indexable)
safe_new_contract('literal', 'int|float|str|bool|tuple|list|None')
for name, tp in inspect.getmembers(ast, inspect.isclass):
    if name[0] == '_':  # python 3.8 added ast._AST which pycontracts does not like
        continue
    safe_new_contract(name, tp)

# Astor tries to get fancy by failing nicely, but in doing so they fail when traversing non-AST type node properties.
#  By deleting this custom handler, it'll fall back to the default ast visit pattern, which skips these missing
# properties. Everything else seems to be implemented, so this will fail silently if it hits an AST node that isn't
# supported but should be.
try:
    del astor.node_util.ExplicitNodeVisitor.visit
except AttributeError:
    # visit isn't defined in this version of astor
    pass


def _pretty_str(o):
    if isinstance(o, ast.AST):
        if isinstance(o, ast.Name):
            return o.id
        elif isinstance(o, ast.Call):
            return "{}(...)".format(_pretty_str(o.func))
        elif isinstance(o, ast.Attribute):
            return "{}.{}".format(_pretty_str(o.value), o.attr)

        return astor.to_source(o).strip()
    else:
        return repr(o)


_log_call_depth = 0


def _log_call(f):
    @functools.wraps(f)
    def inner(*args, **kwargs):
        global _log_call_depth

        result = None
        ex = None
        log.debug("START {}{}({})".format(
            ' ' * _log_call_depth,
            f.__name__,
            ', '.join(
                [_pretty_str(a) for a in args] +
                ["{}={}".format(_pretty_str(k), _pretty_str(v)) for k, v in kwargs.items()]
            )
        ))

        _log_call_depth += 1
        try:
            result = f(*args, **kwargs)
            return result

        except Exception as e:
            ex = e
            raise e
        finally:
            _log_call_depth -= 1

            log.debug("END   {}{}({}) -> {}".format(
                ' ' * _log_call_depth,
                f.__name__,
                ', '.join(
                    [_pretty_str(a) for a in args] +
                    ["{}={}".format(_pretty_str(k), _pretty_str(v)) for k, v in kwargs.items()]
                ),
                _pretty_str(result)
            ), exc_info=ex)

    return inner


from .stack import DictStack
from .resolve import *
from .transformer import *



import ast
import astor
import tempfile
import sys
from miniutils import magic_contract, optional_argument_decorator
from types import ModuleType
import logging
log = logging.getLogger(__name__)

from .core import TrackedContextTransformer, make_function_transformer, resolve_literal
from .core.transformer import function_ast
from .core.resolve import make_ast_from_literal

_exclude = {'__builtin__', '__builtins__', 'builtin', 'builtins'}
if sys.version_info < (3, 6):
    _ast_str_types = ast.Str
else:
    _ast_str_types = (ast.Str, ast.JoinedStr)


@optional_argument_decorator
class lift:
    @magic_contract
    def __init__(self, return_source=False, save_source=True, annotate_types=False, defaults=False, lift_globals=None, imports=True):
        """Converts a closure or method into a pure function which accepts locally defined variables as keyword arguments

        :param return_source: Returns the transformed function's source code instead of compiling it
        :type return_source: bool
        :param save_source: Saves the function source code to a tempfile to make it inspectable
        :type save_source: bool
        :param annotate_types: Flag (or list of var names, or mapping) to use the types of closure variables as the type annotation of the keyword arguments
        :type annotate_types: bool|list|set|tuple|dict
        :param defaults: Default values for free arguments. Must be a dictionary of literals or AST expression, or a bool in which case a best effort is made to convert the closure values into default values
        :type defaults: bool|list|set|tuple|dict
        :param lift_globals: List of global variables to lift to keyword arguments
        :type lift_globals: None|list|set|tuple
        :param imports: Flag or list of imports to include within the function body
        :type imports: bool|list|set|tuple
        """

        self.return_source = return_source
        self.save_source = save_source
        self.annotate_types = annotate_types
        self.defaults = defaults
        self.lift_globals = lift_globals
        self.imports = imports

    def _annotate(self, k, v):
        if self.annotate_types:
            if isinstance(self.annotate_types, bool):
                return ast.Name(id=type(v).__name__, ctx=ast.Load())
            elif isinstance(self.annotate_types, (tuple, list, set)):
                if k in self.annotate_types:
                    return ast.Name(id=type(v).__name__, ctx=ast.Load())
            elif isinstance(self.annotate_types, dict):
                if k in self.annotate_types:
                    result = self.annotate_types[k]
                    if isinstance(result, str):
                        result = ast.Str(s=result)
                    if result and not isinstance(result, ast.expr):
                        raise TypeError("Type annotation must be a string or AST expression (got {})".format(result))
                    return result
        return None

    def _get_default(self, k, v):
        if self.defaults:
            if isinstance(self.defaults, bool):
                attempt = v
            elif isinstance(self.defaults, (tuple, list, set)):
                if k in self.defaults:
                    attempt = v
                else:
                    return None
            elif isinstance(self.defaults, dict):
                if k in self.defaults:
                    attempt = self.defaults[k]
                else:
                    return None
            else:
                return None

            try:
                res = make_ast_from_literal(attempt)
                assert isinstance(res, ast.expr)
                return res
            except (TypeError, AssertionError):
                log.debug("Failed to convert {} to an AST expression".format(attempt))
                return None

        return None

    def _get_free_vars(self, f):
        if isinstance(f.__closure__, tuple):
            free_vars = [(k, v.cell_contents) for k, v in zip(f.__code__.co_freevars, f.__closure__)]
        else:
            free_vars = []

        for glbl in self.lift_globals or []:
            free_vars.append((glbl, f.__globals__[glbl]))

        return free_vars

    def _insert_imports(self, f, f_body, free_vars):
        add_imports = []

        for k, v in f.__globals__.items():
            if isinstance(v, ModuleType):
                add_imports.append((k, v))

        old_free_vars = free_vars
        free_vars = []
        for k, v in old_free_vars:
            if isinstance(v, ModuleType):
                add_imports.append((k, v))
            else:
                free_vars.append((k, v))

        if isinstance(f_body[0], ast.Expr) and isinstance(f_body[0].value, _ast_str_types):
            f_docstring = f_body[:1]
            f_body = f_body[1:]
        else:
            f_docstring = []

        f_body = f_docstring + [
            ast.Import(names=[ast.alias(name=v.__name__, asname=k if k != v.__name__ else None)])
            for k, v in add_imports
            if (isinstance(self.imports, bool) or k in self.imports) and k not in _exclude
        ] + f_body
        return f_body, free_vars

    @magic_contract(f='Callable', returns='Callable|str')
    def __call__(self, f):
        f_mod, f_body, f_file = function_ast(f)
        # Grab function closure variables

        free_vars = self._get_free_vars(f)

        if self.imports:
            f_body, free_vars = self._insert_imports(f, f_body, free_vars)

        func_def = f_mod.body[0]

        new_kws = [ast.arg(arg=k, annotation=self._annotate(k, v)) for k, v in free_vars]
        new_kw_defaults = [self._get_default(k, v) for k, v in free_vars]

        # python 3.8 introduced a new signature for ast.arguments.__init__, so use whatever they use
        ast_arguments_dict = func_def.args.__dict__
        ast_arguments_dict['kwonlyargs'] += new_kws
        ast_arguments_dict['kw_defaults'] += new_kw_defaults

        new_func_def = ast.FunctionDef(
            name=func_def.name,
            body=f_body,
            decorator_list=[],  # func_def.decorator_list,
            returns=func_def.returns,
            args=ast.arguments(**ast_arguments_dict)
        )

        f_mod.body[0] = new_func_def

        if self.return_source or self.save_source:
            try:
                source = astor.to_source(f_mod)
            except ImportError:  # pragma: nocover
                raise ImportError("miniutils.pragma.{name} requires 'astor' to be installed to obtain source code"
                                  .format(name=lift.__name__))
            except Exception as ex:  # pragma: nocover
                raise RuntimeError(astor.dump_tree(f_mod)) from ex
        else:
            source = None

        if self.return_source:
            return source
        else:
            f_mod = ast.fix_missing_locations(f_mod)
            if self.save_source:
                temp = tempfile.NamedTemporaryFile('w', delete=True)
                f_file = temp.name
            no_globals = {}
            exec(compile(f_mod, f_file, 'exec'), no_globals)
            func = no_globals[f_mod.body[0].name]
            if self.save_source:
                func.__tempfile__ = temp
                temp.write(source)
                temp.flush()
            return func



import ast
import copy
import inspect
import sys
import tempfile
import textwrap
import logging
import warnings
log = logging.getLogger(__name__)

import astor
from miniutils.opt_decorator import optional_argument_decorator
from miniutils import magic_contract

from pragma.core.stack import DictStack
from pragma.core.resolve import resolve_literal, resolve_iterable, resolve_indexable, resolve_name_or_attribute, make_ast_from_literal


@magic_contract
def function_ast(f):
    """
    Returns ast for the given function. Gives a tuple of (ast_module, function_body, function_file
    :param f: The function to parse
    :type f: Callable
    :return: The relevant AST code: A module including only the function definition; the func body; the func file
    :rtype: tuple(Module, list(AST), str)
    """
    try:
        f_file = sys.modules[f.__module__].__file__
    except (KeyError, AttributeError):  # pragma: nocover
        f_file = ''

    try:
        found = inspect.findsource(f)
    except IndexError as err:
        raise IOError((
            'Discrepancy in number of decorator @magics expected by '
            'inspect vs. __code__.co_firstlineno\n'
            '{} in {}.\n'
            'Try using the decorators after declaring the function'
            'instead of @-magic').format(f, f_file)
        ) from err

    root = ast.parse(textwrap.dedent(inspect.getsource(f)), f_file)
    return root, root.body[0].body, f_file


class DebugTransformerMixin:  # pragma: nocover
    def visit(self, node):
        orig_node_code = astor.to_source(node).strip()
        log.debug("Starting to visit >> {} << ({})".format(orig_node_code, type(node)))

        new_node = super().visit(node)

        try:
            if new_node is None:
                log.debug("Deleted >>> {} <<<".format(orig_node_code))
            elif isinstance(new_node, ast.AST):
                log.debug("Converted >>> {} <<< to >>> {} <<<".format(orig_node_code, astor.to_source(new_node).strip()))
            elif isinstance(new_node, list):
                log.debug("Converted >>> {} <<< to [[[ {} ]]]".format(orig_node_code, ", ".join(
                    astor.to_source(n).strip() for n in new_node)))
        except Exception as ex:
            log.error("Failed on {} >>> {}".format(orig_node_code, astor.dump_tree(new_node)), exc_info=ex)
            raise ex

        return new_node


class TrackedContextTransformer(DebugTransformerMixin, ast.NodeTransformer):
    def __init__(self, ctxt=None):
        super().__init__()
        self.ctxt = ctxt or DictStack()
        self.conditional_execution = False
        self.in_main_func = False
        self.code_blocks = []

    def visit_many(self, nodes):
        for n in nodes:
            n = self.visit(n)
            if n is None:  # pragma: nocover
                continue
            elif isinstance(n, (list, tuple)):
                yield from n
            else:
                yield n

    def nested_visit(self, nodes, set_conditional_exec=True):
        """When we visit a block of statements, create a new "code block" and push statements into it"""
        was_conditional_exec = self.conditional_execution
        if set_conditional_exec:
            self.conditional_execution = True

        lst = []
        self.code_blocks.append(lst)

        for line in self.visit_many(nodes):
            lst.append(line)

        self.code_blocks.pop()

        self.conditional_execution = was_conditional_exec
        return lst

    def resolve_literal(self, node, raw=False):
        log.debug("Attempting to resolve {} as literal".format(node))
        resolution = resolve_literal(node, self.ctxt, give_raw_result=raw)
        log.debug("Resolved {} to {}".format(node, resolution) if resolution is not node
                  else "Failed to resolve {} as literal".format(node))
        return resolution

    def resolve_name_or_attribute(self, node):
        log.debug("Attempting to resolve name/attr {}".format(node))
        resolution = resolve_name_or_attribute(node, self.ctxt)
        log.debug("Resolved {} to {}".format(node, resolution) if resolution is not node
                  else "Failed to resolve name/attr {}".format(node))
        return resolution

    def resolve_iterable(self, node):
        log.debug("Attempting to resolve {} as iterable".format(node))
        resolution = resolve_iterable(node, self.ctxt)
        log.debug("Resolved {} to {}".format(node, resolution) if resolution is not None
                  else "Failed to resolve {} as iterable".format(node))
        return resolution

    def resolve_indexable(self, node):
        log.debug("Attempting to resolve {} as indexable".format(node))
        resolution = resolve_indexable(node, self.ctxt)
        log.debug("Resolved {} to {}".format(node, resolution) if resolution is not None
                  else "Failed to resolve {} as indexable".format(node))
        return resolution

    def generic_visit_less(self, node, *without):
        for field, old_value in ast.iter_fields(node):
            if field in without:
                continue
            elif isinstance(old_value, list):
                new_values = []
                for value in old_value:
                    if isinstance(value, ast.AST):
                        value = self.visit(value)
                        if value is None:
                            continue
                        elif not isinstance(value, ast.AST):
                            new_values.extend(value)
                            continue
                    new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, ast.AST):
                new_node = self.visit(old_value)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)
        return node

    def __setitem__(self, key, value):
        log.debug("Context setting {} to {}".format(key, "?UNKNOWN?" if value is None else astor.to_source(value).strip()))
        self.ctxt[key] = value

    def __getitem__(self, item):
        res = self.ctxt[item]
        log.debug("Context resolving {} to {}".format(item, "?UNKNOWN?" if res is None else astor.to_source(res).strip()))
        return res

    def __delitem__(self, key):
        log.debug("Context deleting {}".format(key))
        del self.ctxt[key]

    def _assign(self, name, val):
        if isinstance(name, (ast.Tuple, tuple, list)):  # (a, b) = ...
            if isinstance(name, ast.Tuple):
                names = name.elts
            else:
                names = name
            if len(names) == 1:
                yield from self._assign(names[0], val)
            elif val is None or self.conditional_execution:  # Something above us failed, we have no idea what this value is getting assigned
                for subname in names:
                    yield from self._assign(subname, None)
            else:
                iterable_val = self.resolve_iterable(val)
                if iterable_val is not None:  # (a, b) = 1, 2
                    iterable_val = iter(iterable_val)
                    try:
                        for subname in names:
                            if isinstance(subname, ast.Starred):  # (a, _*b_) = 1, _2, 3_
                                yield from self._assign(subname.value, list(iterable_val))
                            else:  # (_a_, b) = _1_, 2
                                yield from self._assign(subname, next(iterable_val))
                    except StopIteration:
                        raise IndexError("Failed to unpack {} into {}, either had too few elements or values after a starred variable".format(val, name))
                else:
                    yield from self._assign(name, None)

        elif isinstance(name, ast.Name):  # a = ...
            if val is None or self.conditional_execution:  # a = ???
                self[name.id] = None
            else:
                literal_val = self.resolve_literal(val)
                if isinstance(literal_val, ast.AST):
                    self[name.id] = literal_val
                else:
                    iterable_val = self.resolve_iterable(val)
                    if iterable_val is not None:
                        self[name.id] = iterable_val
                    else:
                        indexable_val = self.resolve_indexable(val)
                        self[name.id] = indexable_val
            yield name.id

        elif isinstance(name, ast.Attribute):  # a.x = ...
            log.debug("Can't handle assignment to attributes yet")

        elif isinstance(name, ast.Subscript):  # ...[i] = ...
            arr = name.value
            if val is None or self.conditional_execution:  # a[i] = ???
                log.debug("Partial assignment to an iterable is not yet supported, killing entire iterable")
                yield from self._assign(arr, None)
            elif isinstance(arr, ast.Name):  # a[i] = ...
                idx = name.slice
                cur_iterable = self.resolve_iterable(arr)
                if cur_iterable is not None:  # a[i] = ... where a = [1,2,3]
                    literal_idx = self.resolve_literal(idx)
                    if not isinstance(literal_idx, ast.AST):  # a[i] = ... where a = [1,2,3], i =
                        try:
                            cur_iterable[literal_idx] = val
                        except (IndexError, KeyError):
                            raise IndexError("Cannot assigned {arr}[{idx}] = {set} where {arr}={arr_val} and {idx}={idx_val}".format(
                                arr=arr.id, idx=idx, set=val, arr_val=cur_iterable, idx_val=literal_idx
                            ))
                        self[arr.id] = make_ast_from_literal(cur_iterable)
                    else:  # a[???] = val
                        self[arr.id] = None
                else:  # ???[i] = val
                    self[arr.id] = None
                yield arr.id
            else:  # a.x[i] = val, or f(x)[i] = val
                log.debug("Can't handle assignment to subscript of non-variables yet")

        else:  # wtf?
            log.warning("Unhandled assignment of {} to {}".format(val, name))

    def assign(self, name, val):
        try:
            return list(self._assign(name, val))
        except TypeError:
            log.debug("Failed to assign {}={}, rvalue cannot be converted to AST".format(name, val))

    def visit_Assign(self, node):
        node.value = self.visit(node.value)
        self.assign(node.targets, node.value)
        return node

    def visit_AugAssign(self, node):
        node = copy.deepcopy(node)
        node.value = self.visit(node.value)
        new_val = self.resolve_literal(ast.BinOp(op=node.op, left=node.target, right=node.value))
        if not isinstance(new_val, ast.BinOp):
            self.assign(node.target, new_val)
        else:
            self.assign(node.target, None)
        return node

    def _delete(self, node):
        if isinstance(node, ast.Name):
            del self.ctxt[node.id]
        elif isinstance(node, ast.Tuple):
            for n in node.elts:
                self._delete(n)
        elif isinstance(node, (tuple, list)):
            for n in node:
                self._delete(n)
        else:
            warnings.warn("{} doesn't yet know how to handle deleting {}".format(self, node))

    def visit_Delete(self, node):
        self._delete(node.targets)
        return self.generic_visit(node)

    ###################################################
    # From here on down, we just have handlers for every AST node that has a "code block" (stmt*)
    ###################################################

    def visit_FunctionDef(self, node):
        if not self.in_main_func:
            self.ctxt.push({}, False)
            node.body = self.nested_visit(node.body, set_conditional_exec=False)
            self.ctxt.pop()
            return self.generic_visit_less(node, 'body')
        else:
            return node

    def visit_AsyncFunctionDef(self, node):
        return self.visit_FunctionDef(node)

    def visit_ClassDef(self, node):
        # self.ctxt.push({}, False)
        # node.body = self.nested_visit(node.body)
        # self.ctxt.pop()
        # return self.generic_visit_less(node, 'body')
        return node

    def visit_For(self, node):
        node.iter = self.visit(node.iter)
        node.body = self.nested_visit(node.body)
        node.orelse = self.nested_visit(node.orelse)
        return self.generic_visit_less(node, 'body', 'orelse', 'iter')

    def visit_AsyncFor(self, node):
        node.iter = self.visit(node.iter)
        node.body = self.nested_visit(node.body)
        node.orelse = self.nested_visit(node.orelse)
        return self.generic_visit_less(node, 'body', 'orelse', 'iter')

    def visit_While(self, node):
        node.body = self.nested_visit(node.body)
        node.orelse = self.nested_visit(node.orelse)
        return self.generic_visit_less(node, 'body', 'orelse')

    def visit_If(self, node):
        node.test = self.visit(node.test)
        node.body = self.nested_visit(node.body)
        node.orelse = self.nested_visit(node.orelse)
        return self.generic_visit_less(node, 'body', 'orelse', 'test')

    def visit_With(self, node):
        node.body = self.nested_visit(node.body, set_conditional_exec=False)
        return self.generic_visit_less(node, 'body')

    def visit_AsyncWith(self, node):
        node.body = self.nested_visit(node.body, set_conditional_exec=False)
        return self.generic_visit_less(node, 'body')

    def visit_Try(self, node):
        node.body = self.nested_visit(node.body)
        node.orelse = self.nested_visit(node.orelse)
        node.finalbody = self.nested_visit(node.finalbody, set_conditional_exec=False)
        return self.generic_visit_less(node, 'body', 'orelse', 'finalbody')

    def visit_Module(self, node):
        node.body = self.nested_visit(node.body, set_conditional_exec=False)
        return self.generic_visit_less(node, 'body')

    def visit_ExceptHandler(self, node):
        node.body = self.nested_visit(node.body)
        return self.generic_visit_less(node, 'body')


def make_function_transformer(transformer_type, name, description, **transformer_kwargs):
    @optional_argument_decorator
    @magic_contract
    def transform(return_source=False, save_source=True, function_globals=None, **kwargs):
        """
        :param return_source: Returns the transformed function's source code instead of compiling it
        :type return_source: bool
        :param save_source: Saves the function source code to a tempfile to make it inspectable
        :type save_source: bool
        :param function_globals: Overridden global name assignments to use when processing the function
        :type function_globals: dict|None
        :param kwargs: Any other environmental variables to provide during unrolling
        :type kwargs: dict
        :return: The transformed function, or its source code if requested
        :rtype: Callable
        """

        @magic_contract(f='Callable', returns='Callable|str')
        def inner(f):
            f_mod, f_body, f_file = function_ast(f)
            # Grab function globals
            glbls = f.__globals__
            # Grab function closure variables
            if isinstance(f.__closure__, tuple):
                glbls.update({k: v.cell_contents for k, v in zip(f.__code__.co_freevars, f.__closure__)})
            # Apply manual globals override
            if function_globals is not None:
                glbls.update(function_globals)
            # print({k: v for k, v in glbls.items() if k not in globals()})
            trans = transformer_type(DictStack(glbls, kwargs), **transformer_kwargs)
            f_mod.body[0].decorator_list = []
            f_mod = trans.visit(f_mod)
            # print(astor.dump_tree(f_mod))
            if return_source or save_source:
                try:
                    source = astor.to_source(f_mod)
                except Exception as ex:  # pragma: nocover
                    raise RuntimeError(astor.dump_tree(f_mod)) from ex
            else:
                source = None

            if return_source:
                return source
            else:
                f_mod = ast.fix_missing_locations(f_mod)
                if save_source:
                    temp = tempfile.NamedTemporaryFile('w', delete=False)
                    f_file = temp.name
                exec(compile(f_mod, f_file, 'exec'), glbls)
                func = glbls[f_mod.body[0].name]
                if save_source:
                    func.__tempfile__ = temp
                    # When there are other decorators, the co_firstlineno of *some* python distributions gets confused
                    # and thinks they will be there even when they are not written to the file, causing readline overflow
                    # So we put some empty lines to make them align
                    temp.write('\n' * func.__code__.co_firstlineno)
                    temp.write(source)
                    temp.flush()
                    temp.close()
                return func

        return inner

    transform.__name__ = name
    transform.__doc__ = '\n'.join([description, transform.__doc__])
    return transform



import warnings

from .core import *
import logging
log = logging.getLogger(__name__)


def has_break(node):
    for field, value in ast.iter_fields(node):
        if isinstance(value, list):
            for item in value:
                if isinstance(item, (ast.Break, ast.Continue)):
                    return True
                if isinstance(item, ast.AST):
                    if has_break(item):
                        return True
        elif isinstance(value, ast.AST):
            if has_break(value):
                return True
    return False


# noinspection PyPep8Naming
class UnrollTransformer(TrackedContextTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loop_vars = []

    def _names(self, node):
        if isinstance(node, ast.Name):
            yield node.id
        elif isinstance(node, ast.Tuple):
            for elt in node.elts:
                yield from self._names(elt)
        else:
            warnings.warn("Not sure how to handle {} in a for loop target list yet".format(astor.to_source(node).strip()))

    def visit_For(self, node):
        iterable = self.resolve_iterable(node.iter)
        if iterable is None:
            return self.generic_visit(node)

        top_level_break = False
        for n in node.body:
            if isinstance(n, ast.Break):
                top_level_break = True
            # We don't need to check if there's a break in an inner loop, since that doesn't affect this loop
            elif isinstance(n, ast.If) and has_break(n):
                # If there's a conditional break, there's not much we can do about that
                # TODO: If the conditional is resolvable at unroll time, then do so
                return self.generic_visit(node)

        result = []
        # loop_var = node.target.id
        # orig_loop_vars = self.loop_vars
        # print("Unrolling 'for {} in {}'".format(loop_var, list(iterable)))
        for val in iterable:
            # self.ctxt.push({loop_var: val})
            # self.loop_vars = orig_loop_vars | {loop_var}
            # self.ctxt.push()
            # self.visit(ast.Assign(targets=(node.target,), value=make_ast_from_literal(val)))
            try:
                val = make_ast_from_literal(val)
            except TypeError:
                log.debug("Failed to unroll loop, {} failed to convert to AST".format(val))
                return self.generic_visit(node)
            self.loop_vars.append(set(self.assign(node.target, val)))
            for body_node in copy.deepcopy(node.body):
                res = self.visit(body_node)
                if isinstance(res, list):
                    result.extend(res)
                elif res is None:
                    continue
                else:
                    result.append(res)
            # result.extend([self.visit(body_node) for body_node in copy.deepcopy(node.body)])
            # self.ctxt.pop()
            if top_level_break:
                first_result = result
                result = []
                for n in first_result:
                    if isinstance(n, ast.Break):
                        break
                    result.append(n)
                break
        # self.loop_vars = orig_loop_vars
        return result

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load) and self.loop_vars and node.id in set.union(*self.loop_vars):
            if node.id in self.ctxt:
                return self.ctxt[node.id]
            raise NameError("'{}' not defined in context".format(node.id))
        return node


# Unroll literal loops
unroll = make_function_transformer(UnrollTransformer, 'unroll', "Unrolls constant loops in the decorated function")



class DictStack:
    """
    Creates a stack of dictionaries to roughly emulate closures and variable environments
    """

    def __init__(self, *base):
        import builtins
        self.dicts = [dict(builtins.__dict__)] + [dict(d) for d in base]
        self.constants = [True] + [False] * len(base)

    def __iter__(self):
        return (key for dct in self.dicts for key in dct.keys())

    def __setitem__(self, key, value):
        # print("SETTING {} = {}".format(key, value))
        self.dicts[-1][key] = value

    def __getitem__(self, item):
        for dct in self.dicts[::-1]:
            if item in dct:
                if dct[item] is None:
                    raise KeyError("Found '{}', but it was set to an unknown value".format(item))
                return dct[item]
        raise KeyError("Can't find '{}' anywhere in the function's context".format(item))

    def __delitem__(self, item):
        for dct in self.dicts[::-1]:
            if item in dct:
                del dct[item]
                return
        raise KeyError()

    def __contains__(self, item):
        return any(item == key for dct in self.dicts for key in dct.keys())

    def items(self):
        items = []
        for dct in self.dicts[::-1]:
            for k, v in dct.items():
                if k not in items:
                    items.append((k, v))
        return items

    def keys(self):
        return set().union(*[dct.keys() for dct in self.dicts])

    def push(self, dct=None, is_constant=False):
        self.dicts.append(dct or {})
        self.constants.append(is_constant)

    def pop(self):
        self.constants.pop()
        return self.dicts.pop()

    def __repr__(self):
        return "{{...{}...{}}}".format(len(self.dicts) - 1, self.dicts[-2:])



import ast

from miniutils import magic_contract

from .collapse_literals import collapse_literals
import logging
log = logging.getLogger(__name__)


# Directly reference elements of constant list, removing literal indexing into that list within a function
@magic_contract
def deindex(iterable, iterable_name, *args, **kwargs):
    """
    :param iterable: The list to deindex in the target function
    :type iterable: iterable
    :param iterable_name: The list's name (must be unique if deindexing multiple lists)
    :type iterable_name: str
    :param args: Other command line arguments (see :func:`collapse_literals` for documentation)
    :type args: tuple
    :param kwargs: Any other environmental variables to provide during unrolling
    :type kwargs: dict
    :return: The unrolled function, or its source code if requested
    :rtype: Callable
    """

    if hasattr(iterable, 'items'):  # Support dicts and the like
        internal_iterable = {k: '{}_{}'.format(iterable_name, k) for k, val in iterable.items()}
        mapping = {internal_iterable[k]: val for k, val in iterable.items()}
    else:  # Support lists, tuples, and the like
        internal_iterable = {i: '{}_{}'.format(iterable_name, i) for i, val in enumerate(iterable)}
        mapping = {internal_iterable[i]: val for i, val in enumerate(iterable)}

    kwargs[iterable_name] = {k: ast.Name(id=name, ctx=ast.Load()) for k, name in internal_iterable.items()}

    return collapse_literals(*args, function_globals=mapping, **kwargs)



from setuptools import setup

setup(
    name='pragma',
    version='0.2.0',
    packages=['pragma', 'pragma.core'],
    url='https://github.com/scnerd/pypragma',
    license='MIT',
    author='scnerd',
    author_email='scnerd@gmail.com',
    description='Python code transformers that mimic pragma compiler directives',
    long_description=open('README.rst').read(),
    install_requires=[
        'miniutils',
        'astor',
    ]
)



import ast

from .core import TrackedContextTransformer, make_function_transformer, primitive_ast_types
import logging
log = logging.getLogger(__name__)


# noinspection PyPep8Naming
class CollapseTransformer(TrackedContextTransformer):
    def visit_Name(self, node):
        res = self.resolve_literal(node)
        if isinstance(res, primitive_ast_types):
            return res
        return node

    def visit_BinOp(self, node):
        return self.resolve_literal(self.generic_visit(node))

    def visit_UnaryOp(self, node):
        return self.resolve_literal(self.generic_visit(node))

    def visit_BoolOp(self, node):
        return self.resolve_literal(self.generic_visit(node))

    def visit_Compare(self, node):
        return self.resolve_literal(self.generic_visit(node))

    def visit_Subscript(self, node):
        return self.resolve_literal(self.generic_visit(node))

    def visit_Call(self, node):
        node = self.generic_visit(node)
        try:
            return self.resolve_literal(node)
        except (AssertionError, TypeError, KeyError, IndexError):
            return node

    def visit_If(self, node):
        cond = self.resolve_literal(node.test, raw=True)
        # print("Attempting to collapse IF conditioned on {}".format(cond))
        if not isinstance(cond, ast.AST):
            log.debug("Collapsing if condition ({} resolved to {})".format(node.test, cond))
            body = node.body if cond else node.orelse
            result = []
            for subnode in body:
                res = self.visit(subnode)
                if res is None:
                    pass
                elif isinstance(res, list):
                    result += res
                else:
                    result.append(res)
            return result
        else:
            return super().visit_If(node)


# Collapse defined literal values, and operations thereof, where possible
collapse_literals = make_function_transformer(CollapseTransformer, 'collapse_literals',
                                              "Collapses literal expressions in the decorated function into single literals")



from collections import OrderedDict as odict

from .core import *
import logging
log = logging.getLogger(__name__)

# stmt = FunctionDef(identifier name, arguments args,
#                        stmt* body, expr* decorator_list, expr? returns)
#           | AsyncFunctionDef(identifier name, arguments args,
#                              stmt* body, expr* decorator_list, expr? returns)
#
#           | ClassDef(identifier name,
#              expr* bases,
#              keyword* keywords,
#              stmt* body,
#              expr* decorator_list)
#           | Return(expr? value)
#
#           | Delete(expr* targets)
#           | Assign(expr* targets, expr value)
#           | AugAssign(expr target, operator op, expr value)
#           -- 'simple' indicates that we annotate simple name without parens
#           | AnnAssign(expr target, expr annotation, expr? value, int simple)
#
#           -- use 'orelse' because else is a keyword in target languages
#           | For(expr target, expr iter, stmt* body, stmt* orelse)
#           | AsyncFor(expr target, expr iter, stmt* body, stmt* orelse)
#           | While(expr test, stmt* body, stmt* orelse)
#           | If(expr test, stmt* body, stmt* orelse)
#           | With(withitem* items, stmt* body)
#           | AsyncWith(withitem* items, stmt* body)
#
#           | Raise(expr? exc, expr? cause)
#           | Try(stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody)
#           | Assert(expr test, expr? msg)
#
#           | Import(alias* names)
#           | ImportFrom(identifier? module, alias* names, int? level)
#
#           | Global(identifier* names)
#           | Nonlocal(identifier* names)
#           | Expr(expr value)
#           | Pass | Break | Continue
#
#           -- XXX Jython will be different
#           -- col_offset is the byte offset in the utf8 string the parser uses
#           attributes (int lineno, int col_offset)

DICT_FMT = "_{fname}_{n}"


# @magic_contract
def make_name(fname, var, n, ctx=ast.Load):
    """
    Create an AST node to represent the given argument name in the given function
    :param fname: Function name
    :type fname: str
    :param var: Argument name
    :type var: str
    :param ctx: Context of this name (LOAD or STORE)
    :type ctx: Load|Store
    :param n: The number to append to this name (to allow for finite recursion)
    :type n: int
    :param fmt: Name format (if not stored in a dictionary)
    :type fmt: str
    :return: An AST node representing this argument
    :rtype: Subscript|Call
    """
    return ast.Subscript(value=ast.Name(id=DICT_FMT.format(fname=fname, n=n), ctx=ast.Load()),
                         slice=ast.Index(ast.Str(var)),
                         ctx=ctx())


class _InlineBodyTransformer(TrackedContextTransformer):
    def __init__(self, func_name, param_names, n):
        self.func_name = func_name
        # print("Func {} takes parameters {}".format(func_name, param_names))
        self.param_names = param_names
        self.in_break_block = False
        self.n = n
        self.had_return = False
        self.had_yield = False
        super().__init__()

    def visit_Name(self, node):
        # Check if this is a parameter, and hasn't had another value assigned to it
        if node.id in self.param_names:
            # print("Found parameter reference {}".format(node.id))
            if node.id not in self.ctxt:
                # If so, get its value from the argument dictionary
                return make_name(self.func_name, node.id, self.n, ctx=type(getattr(node, 'ctx', ast.Load())))
            else:
                # print("But it's been overwritten to {} = {}".format(node.id, self.ctxt[node.id]))
                pass
        return node

    def visit_Return(self, node):
        if self.in_break_block:
            raise NotImplementedError("miniutils.pragma.inline cannot handle returns from within a loop")
        result = []
        if node.value:
            result.append(ast.Assign(targets=[make_name(self.func_name, 'return', self.n, ctx=ast.Store)],
                                     value=self.visit(node.value)))
        result.append(ast.Break())
        self.had_return = True
        return result

    def visit_Yield(self, node):
        self.had_yield = True
        if node.value:
            return ast.Call(func=ast.Attribute(value=make_name(self.func_name, 'yield', self.n),
                                               attr='append',
                                               ctx=ast.Load()),
                            args=[self.visit(node.value)],
                            keywords=[])
        return node

    def visit_YieldFrom(self, node):
        self.had_yield = True
        return ast.Call(func=ast.Attribute(value=make_name(self.func_name, 'yield', self.n),
                                           attr='extend',
                                           ctx=ast.Load()),
                        args=[self.visit(node.value)],
                        keywords=[])

    def visit_For(self, node):
        orig_in_break_block = self.in_break_block
        self.in_break_block = True
        res = super().visit_For(node)
        self.in_break_block = orig_in_break_block
        return res

    def visit_While(self, node):
        orig_in_break_block = self.in_break_block
        self.in_break_block = True
        res = super().visit_While(node)
        self.in_break_block = orig_in_break_block
        return res


class InlineTransformer(TrackedContextTransformer):
    def __init__(self, *args, funs=None, max_depth=1, **kwargs):
        assert funs is not None
        super().__init__(*args, **kwargs)

        self.funs = funs
        self.max_depth = max_depth

    def visit_Call(self, node):
        """When we see a function call, insert the function body into the current code block, then replace the call
        with the return expression """
        node = self.generic_visit(node)
        node_fun = self.resolve_name_or_attribute(self.resolve_literal(node.func))

        for (fun, fname, fsig, fbody) in self.funs:
            if fun != node_fun:
                continue

            n = 0
            for i in range(self.max_depth):
                args_dict_name = DICT_FMT.format(fname=fname, n=i)
                n = i  # This is redundant, but a bit clearer and safer than just referencing i later
                if args_dict_name not in self.ctxt:
                    break
            else:
                warnings.warn("Inline hit recursion limit, using normal function call")
                return node

            func_for_inlining = _InlineBodyTransformer(fname, fsig.parameters, n)
            fbody = list(func_for_inlining.visit_many(copy.deepcopy(fbody)))

            # print(self.code_blocks)
            cur_block = self.code_blocks[-1]
            new_code = []

            # Load arguments into their appropriate variables
            args = node.args
            flattened_args = []
            for a in args:
                if isinstance(a, ast.Starred):
                    a = self.resolve_iterable(a.value)
                    if a:
                        flattened_args.extend(a)
                    else:
                        warnings.warn("Cannot inline function call that uses non-constant star args")
                        return node
                else:
                    flattened_args.append(a)

            keywords = [(kw.arg, kw.value) for kw in node.keywords if kw.arg is not None]
            kw_dict = [kw.value for kw in node.keywords if kw.arg is None]
            kw_dict = kw_dict[0] if kw_dict else None

            bound_args = fsig.bind(*flattened_args, **odict(keywords))
            bound_args.apply_defaults()

            # Create args dictionary
            final_args = []
            final_kwargs = []

            for arg_name, arg_value in bound_args.arguments.items():
                if isinstance(arg_value, tuple):
                    arg_value = ast.Tuple(elts=list(arg_value), ctx=ast.Load())
                elif isinstance(arg_value, dict):
                    keys, values = zip(*list(arg_value.items()))
                    keys = [ast.Str(k) for k in keys]
                    values = list(values)
                    arg_value = ast.Dict(keys=keys, values=values)
                # fun_name['param_name'] = param_value
                final_kwargs.append((arg_name, arg_value))

            if kw_dict:
                final_kwargs.append((None, kw_dict))

            if func_for_inlining.had_yield:
                final_args.append(ast.List(elts=[ast.Tuple(elts=[ast.Str('yield'), ast.List(elts=[], ctx=ast.Load())],
                                                           ctx=ast.Load())],
                                           ctx=ast.Load()))

            # fun_name = {}
            dict_call = ast.Call(
                func=ast.Name(id='dict', ctx=ast.Load()),
                args=final_args,
                keywords=[ast.keyword(arg=name, value=val) for name, val in final_kwargs]
            )
            new_code.append(ast.Assign(
                targets=[ast.Name(id=args_dict_name, ctx=ast.Store())],
                value=dict_call
            ))

            # Process assignments before resolving body
            cur_block.extend(self.visit_many(new_code))

            # Inline function code
            new_body = list(self.visit_many(fbody))

            # cur_block.append(self.visit(ast.For(target=ast.Name(id='____', ctx=ast.Store()),
            #                                     iter=ast.List(elts=[ast.NameConstant(None)], ctx=ast.Load()),
            #                                     body=new_body,
            #                                     orelse=[])))
            cur_block.append(ast.For(target=ast.Name(id='____', ctx=ast.Store()),
                                     iter=ast.List(elts=[ast.NameConstant(None)], ctx=ast.Load()),
                                     body=new_body,
                                     orelse=[]))

            # fun_name['return']
            if func_for_inlining.had_yield or func_for_inlining.had_return:
                for j in range(100000):
                    output_name = DICT_FMT.format(fname=fname + '_return', n=j)
                    if output_name not in self.ctxt:
                        break
                else:
                    raise RuntimeError("Function {} called and returned too many times during inlining, not able to "
                                       "put the return value into a uniquely named variable".format(fname))

                if func_for_inlining.had_yield:
                    cur_block.append(self.visit(ast.Assign(targets=[ast.Name(id=output_name, ctx=ast.Store())],
                                                           value=make_name(fname, 'yield', n))))
                elif func_for_inlining.had_return:
                    get_call = ast.Call(
                        func=ast.Attribute(
                            value=ast.Name(id=args_dict_name, ctx=ast.Load()),
                            attr='get',
                            ctx=ast.Load()),
                        args=[ast.Str('return'), ast.NameConstant(None)],
                        keywords=[]
                    )
                    cur_block.append(self.visit(ast.Assign(targets=[ast.Name(id=output_name, ctx=ast.Store())],
                                                           value=get_call)))

                return_node = ast.Name(id=output_name, ctx=ast.Load())
            else:
                return_node = ast.NameConstant(None)

            cur_block.append(self.visit(ast.Delete(targets=[ast.Name(id=args_dict_name, ctx=ast.Del())])))
            return return_node

        else:
            return node


# @magic_contract
def inline(*funs_to_inline, max_depth=1, **kwargs):
    """
    :param funs_to_inline: The inner called function that should be inlined in the wrapped function
    :type funs_to_inline: tuple(function)
    :param max_depth: The maximum number of times to inline the provided function (limits recursion)
    :type max_depth: int
    :return: The unrolled function, or its source code if requested
    :rtype: Callable
    """
    funs = []
    for fun_to_inline in funs_to_inline:
        fname = fun_to_inline.__name__
        fsig = inspect.signature(fun_to_inline)
        _, fbody, _ = function_ast(fun_to_inline)

        funs.append((fun_to_inline, fname, fsig, fbody))

    return make_function_transformer(InlineTransformer,
                                     'inline',
                                     'Inline the specified function within the decorated function',
                                     funs=funs, max_depth=max_depth)(**kwargs)
